#include "MeasurementModel.h"


#define PI 3.1415926535897932
using namespace std;

IndependentMVGaussianModel::IndependentMVGaussianModel(int dimension_, vector<double> mean_, vector<double> stdev_) :dimension(dimension_){
    dist.resize(dimension);
    mean = mean_;
    stdev = stdev_;
    for (int i = 0; i < dimension; ++i) {
        dist[i] = normal_distribution<double>(mean[i], stdev[i]);
    }
    unsigned seed1 = std::chrono::system_clock::now().time_since_epoch().count();
    std::mt19937 gen_temp(seed1);
    gen = gen_temp;
}

void IndependentMVGaussianModel::sample(vector<double> & blank_vector){
    //repeatable random numbers
    //std::default_random_engine gen;
    
    //actual random numbers
    //std::random_device rd;
    
    blank_vector.resize(mean.size());
    for (int i = 0; i < mean.size(); ++i) {
        blank_vector[i] = dist[i](gen);
    }
}

//helper functions for computing the Gaussian transition probability
void compute_mean(const vector<Particle *> & particles, Particle & avg_particle){
    
    //overloaded += and /= operators for particles.
    //They act like the addition and scalar division operators for vectors on the Particle's state vector
    for (int i = 0; i < particles.size(); ++i) {
        avg_particle += *(particles[i]);
    }
    avg_particle /= particles.size();
}

void compute_stdev(const vector<Particle *> & particles, Particle & sample_mean_particle, Particle & stdev_particle){
    for (int i = 0; i < particles.size(); ++i) {
        stdev_particle += squared_difference((particles[i])->state, sample_mean_particle.state);
    }
    stdev_particle /= (particles.size());
    for (int i = 0; i < stdev_particle.dimension; ++i) {
        stdev_particle.state[i] = sqrt(stdev_particle.state[i]);
    }
}

void IndependentMVGaussianModel::compute_transition_probability(const vector<Particle *> & input, vector<double> & output){
    
    //compute the sample mean and stdev for the current iteration
    Particle sample_mean_particle(dimension);
    compute_mean(input, sample_mean_particle);

    Particle sample_stdev_particle(dimension);
    compute_stdev(input, sample_mean_particle, sample_stdev_particle);

    //optimized to avoid unnecessary multiplies by using e^a * e^b = e^(a+b)
    for (int j = 0; j < input.size(); ++j) {
        double exp_arg = 0;
        double product_of_stdev = 1;
        // compute the probability of each transition
        for (int i = 0; i < dimension; ++i) {
            exp_arg += pow( ((input[j]->state[i] - sample_mean_particle.state[i])/sample_stdev_particle.state[i]), 2);
            product_of_stdev *= stdev[i];
        }
    output[j] = pow(1 / sqrt(2 * PI), dimension) * 1/product_of_stdev * exp(-0.5 * exp_arg);
    }
}

//compute probability of observation given state
//this model assumes that the PDF of the observation given the model is Gaussian, centered around the model's mean with a stdev that must be given
void IndependentMVGaussianModel::compute_measurement_probability(const vector<Particle *> & input, const Observation & ops, vector<double> & output){
    
    for (int j = 0; j < input.size(); ++j) {
        output[j] = 1;
        double exp_arg = 0;
        double product_of_stdev = 1;
        for (int i = 0; i < dimension; ++i) {
            exp_arg += pow(((ops.state[i] - input[j]->state[i])/stdev[i]), 2);
            product_of_stdev *= stdev[i];
        }
        output[j] = pow(1 / sqrt(2 * PI), dimension) * 1/product_of_stdev * exp(-0.5 * exp_arg);
        //   double factor = 1 / (stdev[i] * sqrt(2 * PI)) * exp(-0.5 * );
        //  output[j] *= factor;
    }
}
