#include "ParticleFilter.h"
#include "math.h"
using namespace std;

ParticleFilter::ParticleFilter(MeasurementModel * measurement_model_, MeasurementModel * transition_model_, int myNumParticles){
    numParticles = myNumParticles;
    particles.resize(numParticles);
    measurement_model = measurement_model_;
    transition_model = transition_model_;
    unsigned seed1 = std::chrono::system_clock::now().time_since_epoch().count();
    std::mt19937 gen_temp(seed1);
    gen = gen_temp;
}

ParticleFilter::ParticleFilter(int numParticles_, int dimension_){
    
    numParticles = numParticles_;
    dimension = dimension_;
    particles.resize(numParticles);
    for (int i = 0; i < numParticles; ++i) {
        Particle * p = new Particle(dimension_);
        particles[i] = p;
    }
    unsigned seed1 = std::chrono::system_clock::now().time_since_epoch().count();
    std::mt19937 gen_temp(seed1);
    gen = gen_temp;
    
}

ParticleFilter::~ParticleFilter(){
    for (int i = 0; i < particles.size(); ++i) {
        delete particles[i];
    };
}

void ParticleFilter::set_measurement_model(MeasurementModel * measurement_model_){
    measurement_model = measurement_model_;
}

void ParticleFilter::set_transition_model(MeasurementModel * transition_model_){
    transition_model = transition_model_;
}

void ParticleFilter::set_resampling_policy(string resampling_policy_, int resampling_interval_, double num_effective_particle_ceiling_){
    resampling_policy = resampling_policy_;
    resampling_interval = resampling_interval_;
    num_effective_particle_ceiling = num_effective_particle_ceiling_;
}

//weighting function for all of the particles
void ParticleFilter::weight_particles(long index_num){
    
    //compute the probability of transitioning from one state to the next using sample statistics
    vector<double> transition_probabilities;
    transition_probabilities.resize(particles.size());
    transition_model->compute_transition_probability(particles, transition_probabilities);
    
    //compute the probability of the observation given the new state
    vector<double> a_post_probabilities;
    a_post_probabilities.resize(particles.size());
    measurement_model->compute_measurement_probability(particles, observations[index_num], a_post_probabilities);
    
    //multiply the two together to get the weights
    vector<double> weights;
    double sum_weights = 0;
    weights.resize(particles.size());
    for (int i = 0; i < a_post_probabilities.size(); ++i) {
        //need to multiply by the current particle weight to allow for different resampling policies
        weights[i] = transition_probabilities[i] * a_post_probabilities[i] * particles[i]->get_weight();
        sum_weights += weights[i];
    }
    if(sum_weights == 0){
        cout << "Model Diverged, increase the number of Particles or alter your model to better fit the data" << endl;
        exit(1);
    }
    //normalize weights to sum to 1
    for (int i = 0; i < weights.size(); ++i) {
        weights[i] /= sum_weights;
        particles[i]->set_weight(weights[i]);
    }
    
}

void ParticleFilter::read_input_data(string filename, double * return_val){
    ifstream noisy_flight_path;
    noisy_flight_path.open (filename);
    
    //header data
    string throwaway;
    getline(noisy_flight_path, throwaway);//first line is title
    
    //to ensure that you don't use a Gaussian noise model on non-gaussian noise
    for (int i = 0; i < 4; ++i) {
        noisy_flight_path >> throwaway;
    }
    if(throwaway != "Gaussian"){
        cout << "Noise Model not Gaussian. Implement a non-gaussian MeasurementModel for optimal performance, exiting." << endl;
        exit(1);
    }
    
    //collect position and velocity noise stdev's
    for (int i = 0; i < 6; ++i) {
        noisy_flight_path >> throwaway;
    }
    double pos_stdev;
    noisy_flight_path >> pos_stdev;
    
    for (int i = 0; i < 6; ++i) {
        noisy_flight_path >> throwaway;
    }
    double vel_stdev;
    noisy_flight_path >> vel_stdev;
    
    //create vector of observations
    double my_t;
    double my_params[4];
    int extract_dt = 0;
    while(noisy_flight_path >> my_t){
        
        if(extract_dt == 1){ //get the dt value from the second time measurement
            dt = my_t;
        }
        noisy_flight_path >> my_params[0];
        noisy_flight_path >> my_params[1];
        noisy_flight_path >> my_params[2];
        noisy_flight_path >> my_params[3];
        observations.emplace_back(my_params, dimension);
        
        ++extract_dt;
    }
    numObservations = observations.size();
    //resize the predicted state vector appropriately for later use
    predicted_states.resize(numObservations);
    
    return_val[0] = pos_stdev;
    return_val[1] = vel_stdev;
}

//initialize particles at the first observation
void ParticleFilter::initialize_particles(){
    //initialize the particles and then form distributions
    for (int i = 0; i < numParticles; ++i) {
        for (int j = 0; j < dimension; ++j) {
            particles[i]->set_state(observations[0].state);
        }
    }
}

//a single time step samples the distributions to move the particles

void ParticleFilter::time_step(){
    
    vector<double> my_sample; //create blank sample;
    vector<double> state;
    for (int i = 0; i < numParticles; ++i) {
        transition_model->sample(my_sample);
        particles[i]->get_state(state); //extract current state
        state[0] += state[2] * dt + my_sample[0]; // x = x + x_velocity * time + random sample
        state[1] += state[3] * dt + my_sample[1]; // y = y + y_velocity * time + random sample
        state[2] += my_sample[2]; //x_velocity += random_sample
        state[3] += my_sample[3]; //y_velocity += random_sample
        particles[i]->set_state(state); //set state
    }
//    for (int i = 0; i < numParticles; ++i) {
//        transition_model->sample(my_sample);
//        particles[i].increment_state_value(particles[i].get_state_value(2) * dt + my_sample[0], 0);
//        particles[i].increment_state_value(particles[i].get_state_value(3) * dt + my_sample[1], 1);
//        particles[i].increment_state_value(my_sample[2], 2);
//        particles[i].increment_state_value(my_sample[3], 3);
//    }
}

void ParticleFilter::set_location_prediction_mode(string mode){
    location_prediction_mode = mode;
}

void ParticleFilter::predict_state(long index_num){
    if(location_prediction_mode == "Mean"){
        vector<double> mean_parameters(dimension, 0);
        //compute the mean value of all the parameters
        for (int i = 0; i < numParticles; ++i) {
            for (int j = 0; j < dimension; ++j) {
                mean_parameters[j] += particles[i]->get_state_value(j) * particles[i]->get_weight();
            }
        }
        //insert mean value into predicted states
        predicted_states[index_num].set_state(mean_parameters);
        
    }
    else if(location_prediction_mode == "Max_Weight"){
        //find the particle with the highest weight
        double max_weight = particles[0]->get_weight();
        int max_index = 0;
        for (int i = 1; i < numParticles; ++i) {
            if(particles[i]->get_weight() > max_weight){
                max_weight = particles[i]->get_weight();
                max_index = i;
            }
        }
        //predict a state using the parameters of highest weighted particle
        vector<double> output = particles[max_index]->get_state_copy();
        predicted_states[index_num].set_state(output);
    }
    else{
        cout << "Error: Prediction Mode Unspecified" << endl;
        exit(1);
    }
}


//start here: generate new particles for each iteration so we don't spiral out of control
void ParticleFilter::resample_particles(){
    //resampling policy logic
    //if using fixed interval mode, only resample ever N times
    if(resampling_policy == "Fixed_Interval"){
        resampling_interval_clock++;
        if(resampling_interval_clock < resampling_interval){
            //don't resample if we're not at the Nth interval
            return;
        }
        else{
            //reset when we get to the Nth interval
            resampling_interval_clock = 0;
        }
    }
    else if(resampling_policy == "Num_Effective_Particles"){
        double num_effective_particles = 0;
        for (int i = 0; i < numParticles; ++i) {
            //sum the squares of the weights
            num_effective_particles += pow(particles[i]->get_weight(),2);
        }
        //effective_num_particles is the inverse of the sum of the squares of the weights
        num_effective_particles = 1 / num_effective_particles;
        //don't resample if the effective number of particles
        if(num_effective_particles < num_effective_particle_ceiling){
            return;
        }
    }
    else{
        cout << "Please do not alter the resampling policy. You specified " <<resampling_policy << ". This is not one of the two valid policies : " << endl;
        cout << "\t 1. Fixed_Interval : resample every N iterations, where N is an integer specified in Sampling_Policy_Parameter"<<endl;
        cout << "\t 2. Num_Effective_Particles : resample when the number of effective particles becomes higher than N, where N is a double specified in Sampling_Policy_Parameter. The effective number of particles e is calculated as e = 1/(sum(w^2)), the inverse of the square sum of weights."<<endl;
        exit(1);
    }
    //repeatable random numbers
    //default_random_engine gen;
    
    //actual random numbers
    vector<double> weights;
    weights.resize(particles.size());
    
    for (int i = 0; i < particles.size(); ++i) {
        weights[i] = particles[i]->get_weight();
    }
    
    discrete_distribution<int> weight_dist {weights.begin(), weights.end()};

    vector<Particle *> new_particles;
    new_particles.resize(particles.size());
    
    for (int i = 0; i < particles.size(); ++i) {
        int index = weight_dist(gen);
        Particle * p = new Particle(*particles[index]);
        new_particles[i] = p;
        new_particles[i]->set_weight(1);//reset the weight when you resample
    }
    
    for (int i = 0; i < particles.size(); ++i) {
        delete particles[i];
        particles[i] = new_particles[i];
    }
}

Observation ParticleFilter::get_prediction(int i){
    return predicted_states[i];
}

void ParticleFilter::print_output(std::string filename){
    //read in observations
    ofstream predicted_path;
    predicted_path.open (filename);
    
    static double time = dt;
    
    for (int i = 0; i < numObservations; ++i) {
        predicted_path << time << " ";
        time += dt;
        for (int j = 0; j < particles[0]->dimension; ++j) {
            predicted_path << predicted_states[i].get_state_value(j) << " ";
        }
        predicted_path << endl;
    }
    
    numObservations = observations.size();
    //resize the predicted state vector appropriately for later use
    predicted_states.resize(numObservations);
    predicted_path.close();
}

vector<double> ParticleFilter::compute_filtered_error(vector<Observation> & true_states, std::string errortype){
    Particle filtered_output(4);
    if(errortype == "MSE"){
        //should this start from 1?
        for (int i = 1; i < true_states.size(); ++i) {
            filtered_output += squared_difference(true_states[i].state, predicted_states[i].state);
        }
        for (int i = 0; i < filtered_output.state.size(); ++i) {
            filtered_output.state[i] /= true_states.size();
        }
    }
    
    else{
        cout << "Error type undefined" << endl;
        exit(1);
    }
    return filtered_output.state;
}

vector<double> ParticleFilter::compute_raw_error(vector<Observation> & true_states, std::string errortype){
    Particle raw_output(4);
    if(errortype == "MSE"){
        //should this start from 1?
        for (int i = 1; i < true_states.size(); ++i) {
            raw_output += squared_difference(true_states[i].state, observations[i].state);
        }
        for (int i = 0; i < raw_output.state.size(); ++i) {
            raw_output.state[i] /= true_states.size();
        }
    }
    
    else{
        cout << "Error type undefined" << endl;
        exit(1);
    }
    return raw_output.state;
}
double ParticleFilter::get_dt(){
    return dt;
}
