#include <iostream>
#include <fstream>
#include "ParticleFilter.h"

using namespace std;

int main(int argc, const char * argv[]) {

    /*cout << "Please specify the file containing the desired simulation parameters " << endl;

    string filename;

    cin >> filename;*/
    
    //simulation parameters are in the config file
    ifstream config_file;
    config_file.open(argv[1]/*filename*/);
    if(!config_file){
        cout << "Please specify a valid file" <<endl;
        exit(1);
    }
    
    //load the data from the config file
    string measured_path_file_name;
    string true_path_file_name;
    string predicted_path_file_name;
    double initial_stdevs[4];
    int num_particles;
    string location_prediction_mode;
    string resampling_policy;
    int interval;
    double num_effective_particles; //should be a double, not necessarily an integer number (misleading name)

    string throwaway; //used to throw away the parameter names in the config file and just extract the values
    config_file >> throwaway >> measured_path_file_name;
    config_file >> throwaway >> true_path_file_name;
    config_file >> throwaway >> predicted_path_file_name;
    config_file >> throwaway >> initial_stdevs[0] >> initial_stdevs[1] >> initial_stdevs[2] >> initial_stdevs[3];
    config_file >> throwaway >> num_particles;
    config_file >> throwaway >> location_prediction_mode;
    config_file >> throwaway >> resampling_policy;
    
    if(resampling_policy == "Fixed_Interval"){
        config_file >> throwaway >> interval;
        num_effective_particles = -1;
    }
    else if(resampling_policy == "Num_Effective_Particles"){
        config_file >> throwaway >> num_effective_particles;
        interval = -1;
    }
    else{
        cout << "Please use a valid resampling policy. You specified " <<resampling_policy << ". This is not one of the two valid policies : " << endl;
        cout << "\t 1. Fixed_Interval : resample every N iterations, where N is an integer specified in Sampling_Policy_Parameter"<<endl;
        cout << "\t 2. Num_Effective_Particles : resample when the number of effective particles becomes higher than N, where N is a double specified in Sampling_Policy_Parameter. The effective number of particles e is calculated as e = 1/(sum(w^2)), the inverse of the square sum of weights." <<endl;
        exit(1);
    }
    
    //Create particle filter object
    ParticleFilter x(num_particles);
    //placeholder for stdevs of measurement model, will be assigned when read from the file in read_input_data
    double params[] = {0,0};
    //read the measurements in
    x.read_input_data(measured_path_file_name, params);
    //initialize the particles from the first observation
    x.initialize_particles();
    
    int dimension = 4; //x, y, x_velocity, y_velocity
    
    //vectors describing noise in each of the 4 dimensions
    vector<double> means; //single vector for means bc we typically use 0-meaned noise
    vector<double> stdev_transition; //transition model's stdev, must be approximated
    vector<double> stdev_measurement; //measurement model's stdev, similar to SNR of measurement system
    
    means.resize(dimension);
    stdev_transition.resize(dimension);
    stdev_measurement.resize(dimension);
    
    //set means to zero bc both models are zero-meaned
    means[0] = 0;
    means[1] = 0;
    means[2] = 0;
    means[3] = 0;
    
    //transition model guesstimated parameters, only used for the first iteration, recomputed using sample mean/stdev every other time
    stdev_transition[0] = initial_stdevs[0];
    stdev_transition[1] = initial_stdevs[1];
    stdev_transition[2] = initial_stdevs[2];
    stdev_transition[3] = initial_stdevs[3];
    
    //state-observation model parameters known from input file
    stdev_measurement[0] = params[0];
    stdev_measurement[1] = params[0];
    stdev_measurement[2] = params[1];
    stdev_measurement[3] = params[1];
    
    //Create Measurement Model Object using Gaussian Model
    IndependentMVGaussianModel measurement_model(dimension, means, stdev_measurement);
    //Need to make a pointer of the object so it can get passed into the Particle Filter
    IndependentMVGaussianModel * measurement_model_pointer = &measurement_model;
    x.set_measurement_model(measurement_model_pointer);
    
    //Create Weighting Model Object using Gaussian Model
    IndependentMVGaussianModel transition_model(dimension, means, stdev_transition);
    //Need to make a pointer of the object so it can get passed into the Particle Filter
    IndependentMVGaussianModel * transition_model_pointer = &transition_model;
    //Create Observation Model Object using Gaussian Model
    x.set_transition_model(transition_model_pointer);
    
    //set the mode for predicting location given the weighted particles
    x.set_location_prediction_mode(location_prediction_mode);
    
    //set the resampling policy
    x.set_resampling_policy(resampling_policy, interval, num_effective_particles);
    
    
    //timing objects
    chrono::time_point<std::chrono::high_resolution_clock> start_time, end_time; // Start and end times
    vector<double> output_times;
    output_times.resize(x.get_numObservations());
    
    //main loop
    //time step the particles forward
    for (int i = 1; i < x.get_numObservations(); ++i) {
        start_time = chrono::high_resolution_clock::now();
        x.time_step();
        //weight particles based off observation
        x.weight_particles(i);
        //predict current state
        x.predict_state(i);
        //generate new particles
        x.resample_particles();
        end_time = chrono::high_resolution_clock::now();
        output_times[i] = (double) std::chrono::duration_cast<std::chrono::microseconds>(end_time - start_time).count();
    }
    //print the predicted path to a file
    x.print_output(predicted_path_file_name);
    //load the true states for comparison
    vector<Observation> true_states;
    load_true_states(true_states, true_path_file_name);
    
    //compare and print the error with and without the particle filter
    vector<double> filtered_error = x.compute_filtered_error(true_states, "MSE");
    vector<double> raw_error = x.compute_raw_error(true_states, "MSE");
    double mean_time = 0;
    double max_time;
    double min_time;
    
    //compute mean_time
    for (int i = 0; i < output_times.size(); ++i) {
        mean_time += output_times[i];
    }
    mean_time /= (output_times.size() - 1);
    //compute minimum and maximum times
    min_time = output_times[1];
    max_time = output_times[1];
    int min_time_idx = 1;
    int max_time_idx = 1;
    for (int i = 2; i < output_times.size(); ++i) {
        if(output_times[i] < min_time){min_time = output_times[i]; min_time_idx = i;}
        if(output_times[i] > max_time){max_time = output_times[i]; max_time_idx = i;}
    }
    
    //compute mean x and y velocities to serve as reference for MSE
    double mag_average_x_vel = 0;
    double mag_average_y_vel = 0;
    for (int i = 0; i < true_states.size(); ++i)
    {
        mag_average_x_vel += abs(true_states[i].state[2]);
        mag_average_y_vel += abs(true_states[i].state[3]);
    }

    mag_average_x_vel /= true_states.size();
    mag_average_y_vel /= true_states.size();

    //output statistics to terminal
    cout << "------- Error Statistics -------" <<endl;
    cout << "--- Before Filtering ---" <<endl;
    cout << "x_pos MSE : " << raw_error[0] << endl;
    cout << "y_pos MSE : " << raw_error[1] << endl;
    cout << "x_vel MSE : " << raw_error[2] << endl;
    cout << "y_vel MSE : " << raw_error[3] << endl;
    cout << "--- After Filtering ---" <<endl;
    cout << "x_pos MSE : " << filtered_error[0] << endl;
    cout << "y_pos MSE : " << filtered_error[1] << endl;
    cout << "x_vel MSE : " << filtered_error[2] << " (average magnitude of x_vel = " << mag_average_x_vel << ")" << endl;
    cout << "y_vel MSE : " << filtered_error[3] << " (average magnitude of y_vel = " << mag_average_y_vel << ")" << endl;
    cout << "------- Time Statistics -------" <<endl;
    cout << "Mean time per iteration: " << mean_time << " usec" << endl;
    cout << "Maximum time over all iterations: " << max_time << " usec" << endl;
    cout << "Minimum time over all iterations: " << min_time << " usec" << endl;
    
    if(max_time < x.get_dt() * 1000000){//dt is in seconds and the times are recorded in usec
        cout << "Particle Filter operated in real-time (maximum time = " << max_time  << " usec < timestep = " << x.get_dt()*1000000  << " usec)"<< endl;
    }
    else{
        cout << "Particle Filter DID NOT operate in real-time (maximum time = " << max_time  << " usec > timestep = " << x.get_dt()*1000000 << " usec)"<< endl;
    }
    cout << "------- End Statistics -------" <<endl;
    
    return 0;
}
